package com.google.inject;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.inject.util.Modules;

import org.junit.Test;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class GenericInjectionTest {

    @Test
    public void testGenericInjection() throws CreationException {
        final List<String> names = Arrays.asList("foo", "bar", "bob");

        Injector injector = Guice.createInjector((Module) new AbstractModule() {
            protected void configure() {
                bind(new TypeLiteral<List<String>>() {
                }).toInstance(names);
            }
        });

        Foo foo = injector.getInstance(Foo.class);
        assertEquals(names, foo.names);
    }

    /**
     * Although we may not have intended to support this behaviour, this test
     * passes under Guice 1.0. The workaround is to add an explicit binding for
     * the parameterized type. See {@link #testExplicitBindingOfGenericType()}.
     */
    @Test
    public void testImplicitBindingOfGenericType() {
        Parameterized<String> parameterized
                = Guice.createInjector().getInstance(Key.get(new TypeLiteral<Parameterized<String>>() {
        }));
        assertNotNull(parameterized);
    }

    @Test
    public void testExplicitBindingOfGenericType() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            protected void configure() {
                bind(Key.get(new TypeLiteral<Parameterized<String>>() {
                }))
                        .to((Class) Parameterized.class);
            }
        });

        Parameterized<String> parameterized
                = injector.getInstance(Key.get(new TypeLiteral<>() {
        }));
        assertNotNull(parameterized);
    }

    @Test
    public void testInjectingParameterizedDependenciesForImplicitBinding() {
        assertParameterizedDepsInjected(new Key<ParameterizedDeps<String, Integer>>() {
                                        },
                Modules.EMPTY_MODULE);
    }

    @Test
    public void testInjectingParameterizedDependenciesForBindingTarget() {
        final TypeLiteral<ParameterizedDeps<String, Integer>> type
                = new TypeLiteral<ParameterizedDeps<String, Integer>>() {
        };

        assertParameterizedDepsInjected(Key.get(Object.class), new AbstractModule() {
            protected void configure() {
                bind(Object.class).to(type);
            }
        });
    }

    @Test
    public void testInjectingParameterizedDependenciesForBindingSource() {
        final TypeLiteral<ParameterizedDeps<String, Integer>> type
                = new TypeLiteral<ParameterizedDeps<String, Integer>>() {
        };

        assertParameterizedDepsInjected(Key.get(type), new AbstractModule() {
            protected void configure() {
                bind(type);
            }
        });
    }

    @Test
    public void testBindingToSubtype() {
        final TypeLiteral<ParameterizedDeps<String, Integer>> type
                = new TypeLiteral<ParameterizedDeps<String, Integer>>() {
        };

        assertParameterizedDepsInjected(Key.get(type), new AbstractModule() {
            protected void configure() {
                bind(type).to(new TypeLiteral<SubParameterizedDeps<String, Long, Integer>>() {
                });
            }
        });
    }

    @Test
    public void testBindingSubtype() {
        final TypeLiteral<SubParameterizedDeps<String, Long, Integer>> type
                = new TypeLiteral<SubParameterizedDeps<String, Long, Integer>>() {
        };

        assertParameterizedDepsInjected(Key.get(type), new AbstractModule() {
            protected void configure() {
                bind(type);
            }
        });
    }

    @SuppressWarnings("unchecked")
    private void assertParameterizedDepsInjected(Key<?> key, Module bindingModule) {
        Module bindDataModule = new AbstractModule() {
            protected void configure() {
            }

            @Provides
            Map<String, Integer> provideMap() {
                return ImmutableMap.of("one", 1, "two", 2);
            }

            @Provides
            Set<String> provideSet(Map<String, Integer> map) {
                return map.keySet();
            }

            @Provides
            Collection<Integer> provideCollection(Map<String, Integer> map) {
                return map.values();
            }
        };

        Injector injector = Guice.createInjector(bindDataModule, bindingModule);
        ParameterizedDeps<String, Integer> parameterizedDeps
                = (ParameterizedDeps<String, Integer>) injector.getInstance(key);
        assertEquals(ImmutableMap.of("one", 1, "two", 2), parameterizedDeps.map);
        assertEquals(ImmutableSet.of("one", "two"), parameterizedDeps.keys);
        assertEquals(ImmutableSet.of(1, 2), ImmutableSet.copyOf(parameterizedDeps.values));
    }

    @Test
    public void testImmediateTypeVariablesAreInjected() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            protected void configure() {
                bind(String.class).toInstance("tee");
            }
        });
        InjectsT<String> injectsT = injector.getInstance(new Key<>() {
        });
        assertEquals("tee", injectsT.t);
    }

    static class Foo {
        @Inject
        List<String> names;
    }

    static class Parameterized<T> {
        @Inject
        Parameterized() {
        }
    }

    static class SubParameterizedDeps<A, B, C> extends ParameterizedDeps<A, C> {
        @Inject
        SubParameterizedDeps(Set<A> keys) {
            super(keys);
        }
    }

    static class ParameterizedDeps<K, V> {
        @Inject
        private Map<K, V> map;
        private Set<K> keys;
        private Collection<V> values;

        @Inject
        ParameterizedDeps(Set<K> keys) {
            this.keys = keys;
        }

        @Inject
        void method(Collection<V> values) {
            this.values = values;
        }
    }

    static class InjectsT<T> {
        @Inject
        T t;
    }
}
